#include "bf.h"

Brainfuck::Brainfuck(const char *filename)
{
    printf("Trainfuck interpreter 0.0.2, (C) Imran Nazar, 2006-2008\n");
    
    FILE *fp;
        
    P=0; IP=0; LP=0; FP=0; oldFP=0;
    sockfd=0; acceptfd=0;
    file = NULL;
    codesize = 0;

    filebuf = (u8*)malloc(256);
    for(int i=0; i<256; i++)
        opcodes[i] = NUL;

    opcodes['>'] = RIGHT;
    opcodes['<'] = LEFT;
    opcodes['+'] = INC;
    opcodes['-'] = DEC;
    opcodes['.'] = OUT;
    opcodes[','] = IN;
    opcodes['['] = DO;
    opcodes[']'] = LOOP;

    opcodes['#'] = FOPEN;
    opcodes[';'] = FREAD;
    opcodes[':'] = FWRITE;
    opcodes['('] = FBACK;
    opcodes[')'] = FFORTH;
    
    opcodes['%'] = CONNECT;
    opcodes['$'] = LISTEN;
    opcodes['@'] = ACCEPT;
    opcodes['`'] = RECV;
    opcodes['\''] = SEND;
    
    data = (u8*)malloc(65536);
    memset(data, 0, 65536);
    
    char **p = environ;
    char *dp = (char*)data;

    while(*p)
    {
        strcpy(dp, *p);
	dp[strlen(*p)] = 255;
	dp += (strlen(*p) + 1);
	p++;
    }

    P = (u8*)dp - data;

    fp = fopen(filename, "r");
    if(!fp)
    {
        printf("Error opening code.\n");
        exit(1);
    }
    
    fseek(fp, 0, SEEK_END);
    int bufsize = ftell(fp);
    
#if BF_DEBUG >= 1
    printf("Reading %d-byte program.\n", bufsize);
#endif
    
    rewind(fp);
    u8 *buf = (u8*)malloc(bufsize);
    u8 c;
    
    do {
        c = fgetc(fp);
        if(opcodes[c] != NUL)
        {
            buf[codesize] = c;
            codesize++;
        }
    } while(!feof(fp));
    
    code = (u8*)malloc(codesize);
    memmove(code, buf, codesize);
    free(buf);

    fclose(fp);

#if BF_DEBUG >= 1
    printf("Code read.\n");
#endif
}

void Brainfuck::run()
{
    while(IP < codesize) exec();
    printf("\nExecution complete.\n");
}

inline void Brainfuck::exec()
{
#if BF_DEBUG >= 2
    printf("%c: ", code[IP]);
#endif
        
    opcodes[code[IP]]();
    IP++;
    
#if BF_DEBUG >= 2
    printf("P=%d, *P=%d, IP=%d\n", P, data[P], IP);
#endif
        
}

void Brainfuck::Error(const char *str)
{
    printf("\nError at instruction %d: %s.\n", IP, str);
    exit(1);
}

void Brainfuck::Dump(int p)
{
    for(int i=0; i<16; i++)
        printf("%02X ", data[p+i]);
    printf("\n");
}

void Brainfuck::NUL() { }

void Brainfuck::RIGHT()
{
    while(code[IP] == '>') { P++; IP++; }
    IP--;
}

void Brainfuck::LEFT()
{
    while(code[IP] == '<') { P--; IP++; }
    IP--;
}

void Brainfuck::INC()
{
    while(code[IP] == '+') { data[P]++; IP++; }
    IP--;
}

void Brainfuck::DEC()
{
    while(code[IP] == '-') { data[P]--; IP++; }
    IP--;
}

void Brainfuck::OUT()
{
    printf("%c", data[P]);
}

void Brainfuck::IN()
{
    // TODO: Keyboard n' such
}

void Brainfuck::DO()
{
    if(data[P] == 0)
        IP = loopstack[LP][1];
    else
    {
        loopstack[LP][0] = IP;
        LP++;
        if(LP > 256) Error("Loop stack overflow");
    }
}

void Brainfuck::LOOP()
{
    LP--;
    loopstack[LP][1] = IP;
    IP = loopstack[LP][0] - 1;
}

void Brainfuck::FOPEN()
{
    if(file)
    {
        fseek(file, (FP+1)&~255, SEEK_SET);
        if(((FP+1)&~255) == (filesize&~255))
        {
            fwrite(filebuf, filesize&255, 1, file);
        }
        else
        {
            fwrite(filebuf, 256, 1, file);
        }
        
        fclose(file);
        
#if BF_DEBUG >= 1
        printf("File closed.\n");
#endif
        
    } else {
        file = fopen((const char*)data+P, "r+");
        if(!file) Error(strerror(errno));
        
        oldFP = 0; FP = 0;
        fseek(file, 0, SEEK_END);
        filesize = ftell(file);
        rewind(file);

        int i = (filesize<256)?filesize:256;
        fread(filebuf, i, 1, file);
        
#if BF_DEBUG >= 1
        printf("File '%s' open.\n", (const char*)data+P);
#endif
        
    }
}

void Brainfuck::FileCheckBuf()
{
    if((FP&~255) != (oldFP&~255))
    {
        fseek(file, (FP+1)&~255, SEEK_SET);
        int i = (filesize<256)?filesize:256;
        fread(filebuf, i, 1, file);
    }
}

void Brainfuck::FFORTH()
{
    oldFP = FP;
    FP++;
    FileCheckBuf();
    data[P] = (FP<filesize)?1:0;
}

void Brainfuck::FBACK()
{
    oldFP = FP;
    FP--;
    FileCheckBuf();
    data[P] = (FP)?1:0;
}

void Brainfuck::FREAD()
{
    data[P] = filebuf[FP&255];
}

void Brainfuck::FWRITE()
{
    filebuf[FP&255] = data[P];
}

void Brainfuck::CONNECT()
{
    if(!sockfd)
    {
        struct sockaddr_in dest;

        int addr = (data[P+0]<<24)+
                   (data[P+1]<<16)+
                   (data[P+2]<<8)+
                   (data[P+3]);
        int port = (data[P+4]<<8)+
                   (data[P+5]);
        
        sockfd = socket(PF_INET, SOCK_STREAM, 0);
        if(sockfd < 0) Error(strerror(errno));

        dest.sin_family      = AF_INET;
        dest.sin_port        = htons(port);
        dest.sin_addr.s_addr = htonl(addr);
        memset(&(dest.sin_zero), 0, 8);

#if BF_DEBUG >= 1
        printf("Connecting to %s port %d.\n", inet_ntoa(dest.sin_addr), port);
#endif
    
        if(connect(sockfd, (struct sockaddr*)&dest, sizeof(struct sockaddr)) < 0)
            Error(strerror(errno));
    
#if BF_DEBUG >= 1
        printf("Connected.\n");
#endif        
    } else {
        if(sockfd > 0)
        {
            close(sockfd);

#if BF_DEBUG >= 1
            printf("Connection closed.\n");
#endif
        }
        sockfd = 0;
    }
}

void Brainfuck::LISTEN()
{
    if(!sockfd)
    {
        struct sockaddr_in dest;

        int addr = (data[P+0]<<24)+
                   (data[P+1]<<16)+
                   (data[P+2]<<8)+
                   (data[P+3]);
        int port = (data[P+4]<<8)+
                   (data[P+5]);
        
        sockfd = socket(PF_INET, SOCK_STREAM, 0);
        if(sockfd < 0) Error(strerror(errno));

        dest.sin_family      = AF_INET;
        dest.sin_port        = htons(port);
        dest.sin_addr.s_addr = htonl(addr);
        memset(&(dest.sin_zero), 0, 8);

        if(bind(sockfd, (struct sockaddr*)&dest, sizeof(struct sockaddr)) < 0)
            Error(strerror(errno));

        if(listen(sockfd, 1) < 0)
            Error(strerror(errno));
    
#if BF_DEBUG >= 1
        printf("Listening on %s port %d.\n", inet_ntoa(dest.sin_addr), port);
#endif
    } else {
        if(sockfd > 0)
        {
            close(sockfd);

#if BF_DEBUG >= 1
            printf("No longer listening.\n");
#endif
        }
        sockfd = 0;
    }
}

void Brainfuck::ACCEPT()
{
    if(!acceptfd)
    {
        struct sockaddr_in src;
        socklen_t src_size = sizeof(struct sockaddr_in);
    
        if((acceptfd = accept(sockfd, (struct sockaddr*)&src, &src_size)) < 0)
            Error(strerror(errno));

#if BF_DEBUG >= 1
        printf("Connection initiated with %s.\n", inet_ntoa(src.sin_addr));
#endif
    } else {
        if(acceptfd > 0)
        {
            close(acceptfd);
        
#if BF_DEBUG >= 1
            printf("Connection closed.\n");
#endif
        }
        acceptfd = 0;
    }
}

void Brainfuck::SEND()
{
    if(acceptfd)
    {
        if(acceptfd != -1)
            if(send(acceptfd, data+P, 1, 0) != 1)
                Error(strerror(errno));
    } else {
        if(sockfd != -1)
            if(send(sockfd, data+P, 1, 0) != 1)
                Error(strerror(errno));
    }
}

void Brainfuck::RECV()
{
    if(acceptfd)
    {
        int i = recv(acceptfd, data+P, 1, 0);
        if(i == -1) Error(strerror(errno));
        else if(i == 0)
        {
            data[P] = 0;
            close(acceptfd);
            acceptfd = -1;
#if BF_DEBUG >= 1
            printf("Connection closed: no more data.\n");
#endif
        }
    } else {
        int i = recv(sockfd, data+P, 1, 0);
        if(i == -1) Error(strerror(errno));
        else if(i == 0)
        {
            data[P] = 0;
            close(sockfd);
            sockfd = -1;
#if BF_DEBUG >= 1
            printf("Connection closed: no more data.\n");
#endif
        }
    }
}

u8 *Brainfuck::code, *Brainfuck::data, *Brainfuck::filebuf;
int Brainfuck::codesize, Brainfuck::filesize;
int Brainfuck::P, Brainfuck::IP, Brainfuck::LP;
int Brainfuck::FP, Brainfuck::oldFP;
int Brainfuck::sockfd, Brainfuck::acceptfd;
FILE *Brainfuck::file;
int Brainfuck::loopstack[256][2];
Brainfuck::fptr Brainfuck::opcodes[256];

